package org.bdware.sc;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.bdware.sc.ContractPort.PortVisitor;
import org.bdware.sc.bean.Contract;
import org.bdware.sc.bean.ContractRequest;
import org.bdware.sc.db.CMTables;
import org.bdware.sc.db.StatusRecorder;
import org.bdware.sc.util.JsonUtil;
import org.bdware.sc.util.LRUList;

import java.io.File;
import java.text.SimpleDateFormat;
import java.util.*;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;

/**
 * @author xiaoche (2021/4/25 18:05)
 */
public class ContractStatusRecorder extends StatusRecorder<ContractMeta> {
    private static final String PREFIX = "CONTRACT_INFO_META_";
    private static final Logger LOGGER = LogManager.getLogger(ContractStatusRecorder.class);
    public static AtomicInteger contractVersion = new AtomicInteger(0);
    // 对外会有一个getContractClient的过程，如果只是想用元信息的，改成getContractClientMeta
    // 如果是想调用的，再使用getContractClient。

    //     合约执行的状态 key:合同id value:合同状态
    //    private static final String DB_NAME = CMTables.ContractInfo.toString();

    static {
        final Object flag = new Object();
        // 调度一个task，在delay（ms）后开始调度，每次调度完后，最少等待period（ms）后才开始调
        ContractManager.scheduledThreadPool.scheduleWithFixedDelay(
                () -> {
                    boolean cleared = dealTimerContractProcess();
                    if (cleared) {
                        synchronized (flag) {
                            try {
                                flag.wait(14000L);
                            } catch (InterruptedException e) {
                                e.printStackTrace();
                            }
                        }
                    }
                },
                0,
                1,
                TimeUnit.SECONDS);
    }

    // contract id to client, changes when some contract is started/stopped/resumed
    private final Map<String, ContractClient> id2ContractClient;

    public LRUList<ContractMeta> runningProcess;

    public ContractStatusRecorder(String dir) {
        super(dir, CMTables.ContractInfo.toString(), PREFIX);
        id2ContractClient = new ConcurrentHashMap<>();
        runningProcess = new LRUList<>();
        LOGGER.info("Load Done!");
        LOGGER.debug("Load Done: " + JsonUtil.toJson(getStatus()));
    }

    public static boolean dealTimerContractProcess() {
        //        if (id2ContractStatus == null) {
        //            String value = readDataFromDB(id2ContractStatusKey);
        //            id2ContractStatus = GSON.fromJson(value, Map.class);
        //            if (id2ContractStatus == null) {
        //                return true;
        //            }
        //            for (Map.Entry<String, ContractStatusEnum> contractStatusEnumEntry :
        // id2ContractStatus.entrySet()) {
        //                //获取合同id
        //                String contractID = contractStatusEnumEntry.getKey();
        //                ContractStatusEnum contractStatusEnum =
        // contractStatusEnumEntry.getValue();
        //                //没有状态信息就停止
        //                if (contractStatusEnum == null) {
        //                    continue;
        //                }
        //                //拥有合同状态信息的，如果是已经执行结束，我们将合同挂起
        //                if
        // (contractStatusEnum.getCode().equals(ContractStatusEnum.STORE.getCode())) {
        //                    resumeContractProcess(contractID);
        //                } else if
        // (contractStatusEnum.getCode().equals(ContractStatusEnum.DONE.getCode())) {
        //                    hangUpContractProcess(contractID);
        //                }
        //            }
        //        }
        return true;
    }

    private static String resumeStartContract(Contract contract) {
        return ContractManager.instance.startContractAndRedirect(contract, null);
    }

    public boolean hasPID(int pid) {
        // TODO 是不是还有别的？这只是在内存里的哦。
        try {
            for (ContractClient c : id2ContractClient.values()) {
                if (Integer.parseInt(c.getPID()) == pid) return true;
            }
        } catch (Exception e) {
            //  e.printStackTrace();
        }
        return false;
    }

    public void reSyncStatusAtStart() {
        List<String> toPrint = new ArrayList<>();
        Set<ContractMeta> toRemove = new HashSet<>();
        for (ContractMeta meta : getStatus().values()) {
            toPrint.add(meta.status + " " + meta.id + " " + meta.name + " ");
            if (meta.contract == null)
                toRemove.add(meta);
        }
        LOGGER.debug(JsonUtil.toPrettyJson(toPrint));
        for (ContractMeta meta : toRemove)
            remove(meta);
        for (String id : id2ContractClient.keySet()) {
            ContractMeta meta = getStatus().get(id);
            if (null == meta) {
                LOGGER.error("unknown id in meta db");
            }
            if (meta != null && meta.status != ContractStatusEnum.RUNNING) {
                meta.status = ContractStatusEnum.RUNNING;
                updateValue(meta);
            }
        }

        for (ContractMeta meta : getStatus().values()) {
            if (meta.status == ContractStatusEnum.RUNNING) {
                if (id2ContractClient.get(meta.getID()) == null) {
                    meta.status = ContractStatusEnum.HANGED;
                    updateValue(meta);
                }
            }
        }
    }

    public void createContract(ContractClient client) {
        ContractMeta meta = client.contractMeta;
        contractVersion.getAndIncrement();
        meta.status = ContractStatusEnum.RUNNING;
        updateValue(meta);
        id2ContractClient.put(meta.id, client);
        runningProcess.add(meta);
    }

    public ContractMeta createIfNotExist(String contractID) {
        ContractMeta ret = getContractMeta(contractID);
        if (null == ret) {
            LOGGER.info("requests don't contain contract " + contractID);
            ret = new ContractMeta();
            ret.id = contractID;
            updateValue(ret);
        }
        return ret;
    }

    public void hangLeastUsedContractProcess() {
        ContractMeta meta = runningProcess.popOldest();
        while (null != meta && ContractStatusEnum.RUNNING != meta.status) {
            meta = runningProcess.popOldest();
        }
        if (null != meta) {
            hangUpContractProcess(meta.id);
        }
    }

    // TODO index name to contract
    public ContractMeta getContractMeta(String idOrNameOrDOI) {
        if (idOrNameOrDOI == null) return null;
        ContractMeta meta = getStatus().get(idOrNameOrDOI);
        if (meta != null) return meta;
        for (ContractMeta cc : getStatus().values()) {
            if (cc.status == ContractStatusEnum.RUNNING || cc.status == ContractStatusEnum.HANGED)
                if (idOrNameOrDOI.equals(cc.name) || idOrNameOrDOI.equals(cc.contract.getDOI())) {
                    return cc;
                }
        }
        for (ContractMeta cc : getStatus().values()) {
            if (idOrNameOrDOI.equals(cc.name) || idOrNameOrDOI.equals(cc.contract.getDOI())) {
                return cc;
            }
        }
        return null;
    }

    public int stopAllContracts() {
        List<String> ids = new ArrayList<>(getStatus().keySet());
        for (String key : ids) {
            killContract(key);
        }
        id2ContractClient.clear();
        return ids.size();
    }

    public String killContract(String idOrName) {
        ContractMeta meta = getContractMeta(idOrName);
        if (null != meta) {
            return killContract(meta);
        }
        return "no such contract " + idOrName;
    }

    public String killContract(ContractMeta meta) {
        contractVersion.getAndIncrement();
        meta.status = ContractStatusEnum.KILLED;
        updateValue(meta);
        ContractClient client = id2ContractClient.get(meta.id);
        if (client != null) {
            client.saveTimesAndTraffic();
            String ff = client.get.syncGet("", "getMemorySet", "");
            LOGGER.info("[killContract] memorySet : " + ff);
            if (null != ff && ff.contains("kill")) {
                // String[] strings = ff.split(";");
                LOGGER.info("need dump when stop");
                String contractName = client.getContractName();
                SimpleDateFormat df = new SimpleDateFormat("yyyy-MM-dd.HH:mm:ss");
                File f =
                        new File(
                                ContractManager.dir + "/memory/" + contractName,
                                df.format(new Date()));
                client.get.syncGet("", "getMemoryDump", f.getAbsolutePath());
                ContractManager.instance.addLocalContractLog("dumpContract", meta);
            }
            ContractManager.instance.invokeContractSuicide(client);
            ContractManager.cPort.updateDb(client.port, false);
        }
        if (null != meta.contractExecutor) {
            meta.contractExecutor.close();
        }
        id2ContractClient.remove(meta.id);
        runningProcess.remove(meta);
        return "success";
    }

    public int stopAllContractsWithOwner(String owner) {
        List<ContractMeta> toKill = new ArrayList<>();
        for (ContractMeta meta : getStatus().values()) {
            if (meta.contract.getOwner().equals(owner)) {
                toKill.add(meta);
            }
        }
        for (ContractMeta meta : toKill) {
            killContract(meta);
        }
        return toKill.size();
    }

    public ContractClient getContractClient(String idOrName) {
        if (idOrName == null || idOrName.length() == 0) {
            return null;
        }
        // TODO ensure load contract client
        // TODO ensure load contract client
        // TODO ensure load contract client
        // TODO ensure load contract client
        ContractMeta meta = getContractMeta(idOrName);
        if (null == meta) {
            LOGGER.error("meta of contract " + idOrName + " not found!");
            return null;
        }
        return id2ContractClient.get(meta.id);
    }

    public void ensureRunning(ContractRequest contractRequest) {
        // 获取合同id
        String contractID = contractRequest.getContractID();
        ContractMeta meta = getContractMeta(contractID);
        contractID = meta.id;
        // 判断目前合同的状态信息
        if (meta.status == ContractStatusEnum.HANGED) {
            resumeContractProcess(contractID);
        }
        // 似乎这样就可以了？
        //        } else if (meta.status == ContractStatusEnum.RUNNING) {
        //            hangUpContractProcess(contractID);
        //        }
    }

    /**
     * 解决思路： 1、CMActions#在executeContract 执行ContractStatusRecorder.
     */
    public synchronized void hangUpContractProcess(String contractID) {
        // dump stop 这里是正常流程代码
        //        ContractClient contractClient = getContractClientById(contractID);
        //        if (contractClient == null) {
        //            return;
        //        }
        //        //挂起
        //        Contract contract = contractClient.contract;
        //        ContractManager.instance.dumpContract(contract.getID(), null);
        //        setContractClient(contractID, contractClient);
        //        ContractManager.instance.hangUpStopContract(contract.getID());
        // todo 临时想强制测试下效果
        ContractMeta contractMeta = getContractMeta(contractID);
        // 挂起
        Contract contract = contractMeta.contract;
        // System.out.println("hangUp"+contract.Mask);
        if (contractMeta.status == ContractStatusEnum.RUNNING) {
            contractMeta.status = ContractStatusEnum.HANGED;
            updateValue(contractMeta);
            ContractClient client = id2ContractClient.get(contractMeta.id);
            if (null != client) {
                client.saveTimesAndTraffic();

                SimpleDateFormat df = new SimpleDateFormat("yyyy-MM-dd.HH_mm_ss"); // 设置日期格式
                File f =
                        new File(
                                ContractManager.dir + "/memory/" + contractMeta.name,
                                df.format(new Date()));
                File parent = f.getParentFile();
                if (!parent.exists()) {
                    LOGGER.trace("make directory " + parent.getAbsolutePath() + ":" + parent.mkdirs());
                }
                ContractManager.instance.dumpContract(contract.getID(), f.getAbsolutePath());
                ContractManager.instance.invokeContractSuicide(client);
                id2ContractClient.remove(contractMeta.getID());
                runningProcess.remove(contractMeta);
            }
        }
    }

    public synchronized void resumeContractProcess(String contractID) {
        // 从参数获取合同名称
        ContractMeta meta = getContractMeta(contractID);
        // 启动合约
        // 加载所需要的内存
        // 该方法调用方法已经考虑了内存问题，但是内存不足的时候，重试或者等待，或者将合约压到失败中稍后重试，可以尝试设计消息重试机制
        // 如果是内存中则唤起合同信息 todo 内存不足  关闭合约追赶，补偿数据
        if (meta.status == ContractStatusEnum.HANGED || meta.status == ContractStatusEnum.KILLED) {
            ContractStatusEnum preStatus = meta.status;
            // TODO Resume里复用的是start逻辑，启动完了就成Running了哦。可能不能这么来。
            String startContractResult = resumeStartContract(meta.contract);
            ContractResult contractResult =
                    JsonUtil.fromJson(startContractResult, ContractResult.class);
            if (contractResult.status == ContractResult.Status.Error) {
                // 记录错误日志
                ContractManager.instance.addLocalContractLog(
                        "resumeContract error",
                        meta.contract.getID(),
                        meta.name,
                        meta.contract.getOwner());
            } else {
                ContractManager.instance.addLocalContractLog(
                        "resumeContract success",
                        meta.contract.getID(),
                        meta.name,
                        meta.contract.getOwner());
                // TODO 可能会重复load相同的镜像？
                // 如果是killed 只根据manifest去判断是否加载memory
                // 可增加一个判断，如果hanged朋manifest里是加载memory，这里就不再加载了。
                ContractClient client = id2ContractClient.get(meta.getID());
                if (preStatus == ContractStatusEnum.HANGED) {

                    ContractManager.instance.loadMemory(
                            client, ContractManager.instance.findNewestMemory(meta.name));


                }
                client.contractMeta.setStatus(ContractStatusEnum.RUNNING);
                updateValue(client.contractMeta);
                // 还要判断一发，是不是已有permission的值。如果有，要setPermission一把。
                //                meta.status = ContractStatusEnum.RUNNING;
                //                updateValue(meta);
            }
        }
    }

    public PortVisitor getVisitor() {
        return new ReconnectVisitor();
    }

    class ReconnectVisitor implements PortVisitor {
        private final Set<Integer> existed = new HashSet<>(); // 已经连上的合约端口
        private final Set<String> contractIDs = new HashSet<>(); // 已经连上的合约id

        public ReconnectVisitor() {
            for (ContractClient c : id2ContractClient.values()) {
                existed.add(c.port);
            }
        }

        @Override
        public boolean visit(int i) {
            if (i == ContractManager.cPort.getCMPort()) {
                return false;
            }
            if (existed.contains(i)) {
                return true;
            }
            try {
                ContractClient client = new ContractClient(i);
                // LOGGER.info("[CM重连] position-----1");
                if (client.isRunning) {
                    LOGGER.debug("CP listened to port " + i + " is running");
                    if (contractIDs.contains(client.getContractID())) {
                        LOGGER.info("find contract process with duplicate contractID, kill!");
                        ContractManager.instance.invokeContractSuicide(client);
                        return false;
                    }
                    contractIDs.add(client.getContractID());
                    if (client.contractMeta.name.equals("analysis_client")) {
                        LOGGER.info("kill analysis_client");
                        ContractManager.instance.invokeContractSuicide(client);
                        return false;
                    }
                    LOGGER.debug(
                            String.format(
                                    "CP listened to port %d:\n\tID=%s\n\tName=%s\n\tType=%s\n"
                                            + "\tKey=%s\n\tPubKey=%s\n\tCopies=%d\n\t%s",
                                    i,
                                    client.getContractID(),
                                    client.contractMeta.name,
                                    client.getContractType(),
                                    client.getContractKey(),
                                    client.getPubkey(),
                                    client.getContractCopies(),
                                    client.contractMeta.contract.startInfo));
                    client.get.syncGet("", "setDir", ContractManager.dir);
                    //                    String str = client.getIdentifier();
                    client.get.syncGet("", "getDumpPeriod", "a");
                    client.get.syncGet("", "startAutoDump", "a");
                    createContract(client);
                    LOGGER.info(
                            String.format("reconnect to port %d: contract %s",
                                    i, client.contractMeta.name));
                    return true;
                }
            } catch (Exception e) {
                //just ignore
            }
            return false;
        }
    }
}
