package org.bdware.sc.event;

import org.bdware.sc.util.DHTUtil;
import org.bdware.sc.util.JsonUtil;

import static org.bdware.sc.ContractManager.instance;

/**
 * @author Kaidong Wu
 */
public class EventCenter {
//    private static final Logger LOGGER = LogManager.getLogger(EventCenter.class);

    /**
     * get the nearest node to the topic in the hash function range
     *
     * @param topic the topic
     * @param k     the number of centers
     * @return id of the node
     */
    public String[] getCenterByTopic(String topic, int k) {
        if (null == instance.nodeCenterConn) {
            return null;
        }
        String[] centers = DHTUtil.getClusterByKey(topic, k);
        if (null == centers || centers.length == 0) {
            return null;
        }
        return centers;
    }

    /**
     * subscribe a topic in center
     *
     * @param topic     event topic
     * @param semantics event semantics, used to mark PRESUB events
     * @param center    id of event center if the subscribing has been handled
     */
    public void subInCenter(String topic, REvent.REventSemantics semantics, String center) {
        if (null == instance.nodeCenterConn) {
            return;
        }
        REvent msg = new REvent(topic,
                REvent.REventType.SUBSCRIBE,
                String.format("{\"subscriber\":\"%s\"}",
                        instance.nodeCenterConn.getNodeId()),
                "");
        msg.setSemantics(semantics);
        msg.doSignature(instance.nodeCenterConn.getNodeKeyPair());
        msg.setCenter(center);
        String[] centers = getCenterByTopic(topic, 2);
        if (null == centers) {
            return;
        }
        if (null != center) {
            if (centers[0].equals(center) && centers.length > 1) {
                instance.masterStub.deliverEvent(centers[1], JsonUtil.toJson(msg));
            } else {
                instance.masterStub.deliverEvent(centers[0], JsonUtil.toJson(msg));
            }
        } else if (!instance.masterStub.deliverEvent(centers[0], JsonUtil.toJson(msg)) && centers.length > 1) {
            msg.setCenter(centers[0]);
            instance.masterStub.deliverEvent(centers[1], JsonUtil.toJson(msg));
        }
    }

    /**
     * deliver an event to center
     *
     * @param topic event topic
     * @param event the event
     * @return if this node is the center, return false; otherwise, return true
     */
    public boolean deliverEvent(String topic, REvent event) {
        if (null == instance.masterStub) {
            return false;
        }
        String[] centers = getCenterByTopic(topic, 1);
        if (null == centers) {
            return false;
        }
        return instance.masterStub.deliverEvent(centers[0], JsonUtil.toJson(event));
    }

    /**
     * publish an event to another node; used by NodeConsumer
     *
     * @param target id of the target node
     * @param eStr   event string
     */
    public void publishEvent(String target, String eStr) {
        if (null != instance.masterStub) {
            instance.masterStub.deliverEvent(target, eStr);
        }
    }
}