package org.bdware.sc.event.clients;

import com.google.gson.JsonObject;
import com.google.gson.annotations.Expose;
import io.netty.channel.Channel;
import io.netty.handler.codec.http.websocketx.TextWebSocketFrame;
import org.bdware.sc.conn.ResultCallback;
import org.bdware.sc.util.JsonUtil;

/**
 * @author Kaidong Wu
 */
public class WSClientConsumer implements IEventConsumer {
    @Expose(serialize = false, deserialize = false)
    private final Channel channel;
    @SuppressWarnings("unused")
    private final ConsumerType type = ConsumerType.WSClient;

    public WSClientConsumer(Channel channel) {
        this.channel = channel;
    }

    public Channel getChannel() {
        return channel;
    }

    @Override
    public String getId() {
        return "WSC_" + channel.id().asLongText();
    }

    @Override
    public ConsumerType getType() {
        return type;
    }

    @Override
    public void publishEvent(Object msg, ResultCallback rc, String... options) {
        if (!channel.isActive()) {
            rc.onResult((String) null);
        }
        JsonObject ret = new JsonObject();
        ret.addProperty("action", "onEvent");
        ret.addProperty("status", "Success");
        ret.add("data", JsonUtil.parseString(msg.toString()));
        channel.writeAndFlush(new TextWebSocketFrame(ret.toString()));
    }

    @Override
    public void competeSub(Object msg, ResultCallback rc, String... options) {
    }
}
