package org.bdware.sc.util;

import org.bdware.sc.ContractManager;

import java.math.BigInteger;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

public class DHTUtil {
    private static final Map<String, BigInteger> ID_INTEGER_CACHE = new HashMap<>();

    public static String[] getClusterByKey(String key, int k) {
        return getClusterByKey(key, k, ContractManager.instance.masterStub.listNodes());
    }

    protected static String[] getClusterByKey(String key, int k, String[] nodes) {
        if (nodes.length == 0) {
            return null;
        }
        if (nodes.length == 1) {
            return nodes;
        }

        String hash = HashUtil.sha3ToFixedLen(key, nodes[0].length() - 2);

        int l = 0, r = nodes.length - 1, m,
                h2l = hash.compareTo(nodes[l].substring(2)), r2h = nodes[r].substring(2).compareTo(hash),
                h2m;
        BigInteger bigH = null;
        String selected;
        do {
            if (h2l < 1) { // if the left is bigger than hash
                selected = nodes[l];
                m = l;
                break;
            }
            if (r2h < 1) { // if the right is smaller than hash
                selected = nodes[r];
                m = r;
                break;
            }
            if (l + 1 == r) {
                BigInteger bigL = getBigInteger(nodes[l]),
                        bigR = getBigInteger(nodes[r]);
                bigH = new BigInteger(hash, 16);
                if (bigR.subtract(bigH).compareTo(bigH.subtract(bigL)) > -1) {
                    selected = nodes[l];
                    m = l;
                } else {
                    selected = nodes[r];
                    m = r;
                }
                break;
            }
            m = (l + r) >> 1;
            h2m = hash.compareTo(nodes[m].substring(2));
            if (h2m < 1) {
                r = m;
                r2h = -h2m;
            } else {
                l = m;
                h2l = h2m;
            }
        } while (true);
        if (k == 1) {
            return new String[]{selected};
        }
        List<String> ret = new ArrayList<>();
        ret.add(selected);
        if (k > 1) {
            l = m - 1;
            r = m + 1;
            while (ret.size() < k && (l >= 0 || r < nodes.length)) {
                if (l < 0) {
                    ret.add(nodes[r++]);
                } else if (r >= nodes.length) {
                    ret.add(nodes[l--]);
                } else {
                    if (null == bigH) {
                        bigH = new BigInteger(hash, 16);
                    }
                    if (getBigInteger(nodes[r]).subtract(bigH)
                            .compareTo(bigH.subtract(getBigInteger(nodes[l]))) > -1) {
                        ret.add(nodes[l--]);
                    } else {
                        ret.add(nodes[r++]);
                    }
                }
            }
        }
        return ret.toArray(new String[0]);
    }

    private static BigInteger getBigInteger(String id) {
        String bigInt = id.substring(2);
        if (!ID_INTEGER_CACHE.containsKey(bigInt)) {
            ID_INTEGER_CACHE.put(bigInt, new BigInteger(bigInt, 16));
        }
        return ID_INTEGER_CACHE.get(bigInt);
    }
}
